
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include "pac7311-stream.h"
#include "pac7311-grab.h"
#include "pac7311-grab-cmd-handlers.h"



void free_cmd_exec_list(cmd_exec_type *exec_list)
{
	cmd_exec_type *next;
	
	while (exec_list)
	{
		next = exec_list->next;
		free(exec_list);
		exec_list = next;
	}
}


int skip_whitespaces(char *cmd_line, int parse_pos)
{
	int i = parse_pos;
	
	while (cmd_line[i] &&
	       ((cmd_line[i] == ' ') ||
	        (cmd_line[i] == '\t')))
	{
		++i;
	}
	
	return i;
}


int skip_token(char *cmd_line, int parse_pos)
{
	int i = parse_pos;
	
	if (cmd_line[i] == '\"')
	{
		char *pos = index(cmd_line + i + 1, '\"');
		
		if (pos == NULL)
		{
			return -1;
		}
		else
		{
			return (int) (pos - cmd_line) + 1;
		}
	}
	
	while (cmd_line[i] &&
	       (cmd_line[i] != ' ') &&
	       (cmd_line[i] != '\t') &&
	       (cmd_line[i] != '\n'))
	{
		++i;
	}
	
	return i;
}


int find_cmd_index(char *cmd)
{
	int i = 0;
	
	while (command_list[i].name != CMD_END)
	{
		if (!strcmp(cmd, command_list[i].name))
		{
			break;
		}
		
		++i;
	}
	
	if (command_list[i].name == CMD_END)
	{
		return -1;
	}
	else
	{
		return i;
	}
}


int parse_hex_str(char *hexseq, unsigned char *converted)
{
	int hs_len, hs_cnt, conv_cnt, high_bit;

	hs_len = strlen(hexseq);

	for (hs_cnt = (hs_len - 1), conv_cnt = 0, high_bit = 0;
	     hs_cnt >= 0;
	     --hs_cnt, high_bit = !high_bit)
	{
		char c = hexseq[hs_cnt];

		if ((c >= 48) && (c <= 57)) /* 0-9 */
		{
			c -= 48;
		}
		else if ((c >= 97) && (c <= 103)) /* a-f */
		{
			c -= 87;
		}
		else if ((c >= 65) && (c <= 70)) /* A-F */
		{
			c -= 55;
		}
		else
		{
			return 1;
		}

		conv_cnt = (hs_cnt / 2);

		if (high_bit)
		{
			converted[conv_cnt] += (c << 4);
		}
		else
		{
			converted[conv_cnt] = c;
		}
	}
	
	return 0;
}


int parse_arg(char *parse_str, arg_type_type arg_type, arg_value_type *arg_val)
{
	unsigned long ulong;
	char *end_ptr;

	switch (arg_type)
	{
	case CMD_ARG_UCHAR:
	case CMD_ARG_ULONG:
	
		ulong = strtoul(parse_str, &end_ptr, 0);
		
		if (*end_ptr)
		{
			return 1;
		}
		
		if (arg_type == CMD_ARG_UCHAR)
		{
			if (ulong > 255)
				return 2;
				
			arg_val->uchar = (unsigned char) ulong;
		}
		else
		{
			arg_val->ulong = ulong;
		}
		
		break;
		
	case CMD_ARG_STR:
	
		if ((parse_str[0] != '\"') ||
		    (parse_str[strlen(parse_str) - 1] != '\"'))
		{
			return 4;
		}
		
		strncpy(arg_val->str, parse_str + 1, strlen(parse_str) - 2);
		arg_val->str[strlen(parse_str) - 2] = '\0';
		break;
	
	case CMD_ARG_HEX_STR:
	
		if (parse_hex_str(parse_str, arg_val->hex_str))
		{
			return 5;
		}
	
		break;
		
	default:
		return 6;
	}

	return 0;
}


int build_cmd_exec_list(FILE *grab_file, cmd_exec_type **pexec_list)
{

#define PARSE_ERROR(ERRNO, MSG, ARGS ...) \
do \
{ \
	fprintf(stderr, "Error " #ERRNO " in line %d: " MSG "\n", line, ##ARGS); \
	retval = ERRNO; \
	goto error; \
} \
while (0);

#define PARSE_TOKEN(cmd_line, parse_pos, parse_str, token_end_pos) \
do \
{ \
	token_end_pos = skip_token(cmd_line, parse_pos); \
	\
	if (token_end_pos < 0) \
	{ \
		PARSE_ERROR(2, "unterminated string"); \
	} \
	\
	strncpy(parse_str, cmd_line + parse_pos, token_end_pos - parse_pos); \
	parse_str[token_end_pos - parse_pos] = '\0'; \
} \
while (0);


	int retval = 0, arg_cnt;

	char cmd_line[MAX_CMD_LINE_SIZE], parse_str[MAX_CMD_LINE_SIZE];
	cmd_exec_type *cur_cmd = *pexec_list = NULL;
	arg_type_type cur_arg_type;
	arg_value_type cur_arg_val;

	int parse_pos, token_end_pos, line = 0, cur_cmd_index;
	int in_comment = 0;
	

	while (fgets(cmd_line, MAX_CMD_LINE_SIZE, grab_file) != NULL) {
	
		++line;
	
		parse_pos = skip_whitespaces(cmd_line, 0);
		
		switch (cmd_line[parse_pos])
		{
		case 0:
			PARSE_ERROR(1, "line too long");
		
		case '#':
			switch (cmd_line[parse_pos + 1])
			{
			case '{':
				in_comment = 1;
				continue;
			case '}':
				in_comment = 0;
				continue;
			}
		case '\n':
			continue;
		}
		
		if (in_comment)
		{
			continue;
		}
		
		PARSE_TOKEN(cmd_line, parse_pos, parse_str, token_end_pos);
		
		if (*pexec_list == NULL)
		{
			cur_cmd = *pexec_list = (cmd_exec_type *) malloc(sizeof(cmd_exec_type));
		}
		else
		{
			cur_cmd->next = (cmd_exec_type *) malloc(sizeof(cmd_exec_type));
			cur_cmd = cur_cmd->next;
		}

		cur_cmd->next = NULL;


		cur_cmd_index = find_cmd_index(parse_str);
		
		if (cur_cmd_index < 0)
		{
			PARSE_ERROR(3, "command not supported (%s)", parse_str);
		}

		cur_cmd->handler = command_list[cur_cmd_index].handler;
		cur_cmd->line = line;

		arg_cnt = 0;
		
		while ((cur_arg_type = command_list[cur_cmd_index].arg_types[arg_cnt])
			!= CMD_ARG_END)
		{
			parse_pos = skip_whitespaces(cmd_line, token_end_pos);

			switch (cmd_line[parse_pos])
			{
			case 0:
				PARSE_ERROR(4, "line too long (parsing arg no %d)", arg_cnt + 1);

			case '\n':
				PARSE_ERROR(5, "not enough arguments supplied for command %s",
				            command_list[cur_cmd_index].name);
			}

			PARSE_TOKEN(cmd_line, parse_pos, parse_str, token_end_pos);
						
			if (parse_arg(parse_str, cur_arg_type, &cur_arg_val))
			{
				PARSE_ERROR(6, "argument %d has wrong type for command %s",
				            arg_cnt + 1, command_list[cur_cmd_index].name);
			}
			
			cur_cmd->arg_values[arg_cnt] = cur_arg_val;
			
			++arg_cnt;
		}

		parse_pos = skip_whitespaces(cmd_line, token_end_pos);
		
		if (cmd_line[parse_pos] == 0)
		{
			PARSE_ERROR(7, "line too long");
		}

		if (cmd_line[parse_pos] != '\n')
		{
			PARSE_ERROR(8, "too many arguments for command %s",
			            command_list[cur_cmd_index].name);
		}
	}
	
	goto exit;

error:
	free_cmd_exec_list(*pexec_list);

exit:
	return retval;
}



int run_cmd_exec_list(cmd_exec_type *exec_list, int devfd)
{
	int ret, retval;

	retval = 0;

	while (exec_list)
	{
		ret = exec_list->handler(devfd, exec_list->arg_values);
		
		if (ret)
		{
			++retval;
			printf("Warning: execution of command in line %d failed, error code %d\n", 
			       exec_list->line, ret);
		}
		
		exec_list = exec_list->next;
	}

	return retval;
}


int main(int argc, char **argv)
{
	FILE *grab_file;
	int devfd, run_errors;
	char *device_name, *grab_file_name;
	cmd_exec_type *exec_list;
	
	if (argc != 3)
	{
		fprintf(stderr, "usage: %s <device> <grab file>\n", argv[0]);
		goto exit0;
	}
	
	device_name = argv[1];
	grab_file_name = argv[2];
	
	if ((devfd = open(device_name, O_RDONLY)) < 0)
	{
		fprintf(stderr, "%s: could not open pac7311 device '%s'\n", argv[0], device_name);
		goto exit0;
	}
	
	if (!(grab_file = fopen(grab_file_name, "r")))
	{
		fprintf(stderr, "%s: could not open grab file '%s'\n", argv[0], grab_file_name);
		goto exit1;
	}

	if (build_cmd_exec_list(grab_file, &exec_list))
	{
		fprintf(stderr, "%s: error while parsing grab file, aborting\n", argv[0]);
		goto exit2;
	}

	run_errors = run_cmd_exec_list(exec_list, devfd);
	
	if (run_errors)
	{
		fprintf(stderr, "%s: run completed, %d commands failed\n", argv[0], run_errors);
		goto exit3;
	}
	
exit3:
	free_cmd_exec_list(exec_list);

exit2:
	fclose(grab_file);

exit1:
	close(devfd);
	
exit0:
	return 0;
}
