

#include "usb-pac7311-ringbuf.h"

#define MIN(A, B) (((A) > (B))? (B):(A))


/******************************************************************************
 *
 * Ringbuffer related stuff
 *
 ******************************************************************************/

static inline size_t ringbuf_init(struct usb_pac7311_ringbuf **ringbuf_p, size_t size)
{
	struct usb_pac7311_ringbuf *ringbuf;

	if (!(ringbuf = *ringbuf_p = kmalloc(sizeof(struct usb_pac7311_ringbuf), GFP_KERNEL)))
		return 0;


	rwlock_init(&(ringbuf->lock));

	ringbuf->size = 0;

	if (!(ringbuf->buffer = kmalloc(size, GFP_KERNEL)))
	{
		kfree(ringbuf);
		return 0;
	}
	
	ringbuf->size = size;
	
	ringbuf->start = ringbuf->end = 0;

	return ringbuf->size;
}

static inline void ringbuf_destroy(struct usb_pac7311_ringbuf *ringbuf)
{
	ringbuf->size = 0;

	kfree(ringbuf->buffer);
	kfree(ringbuf);
}

static inline void ringbuf_reset(struct usb_pac7311_ringbuf *ringbuf)
{
	unsigned long flags;

	write_lock_irqsave(&(ringbuf->lock), flags);

	ringbuf->start = ringbuf->end = 0;

	write_unlock_irqrestore(&(ringbuf->lock), flags);
}


/* ringbuf_data_size : size of the data in the ringbuffer
 *
 * EEEE>DDDD<EEEE  : end - start
 * DDDD<EEEE>DDDD  : size + end - start
 */
static inline size_t __ringbuf_data_size(struct usb_pac7311_ringbuf *ringbuf, int lock)
{
	unsigned long flags;

	size_t size;

	if (lock)
	{
		read_lock_irqsave(&(ringbuf->lock), flags);
	}
	
	size = ((ringbuf->end >= ringbuf->start)?
	        (ringbuf->end - ringbuf->start) :
	        (ringbuf->size + ringbuf->end - ringbuf->start));
	                    

	if (lock)
	{
		read_unlock_irqrestore(&(ringbuf->lock), flags);
	}
	
	return size;
}


static inline size_t ringbuf_data_size(struct usb_pac7311_ringbuf *ringbuf)
{
	return __ringbuf_data_size(ringbuf, 1);
}


static size_t ringbuf_data_write(struct usb_pac7311_ringbuf *ringbuf, unsigned char *data, size_t count)
{
	unsigned long flags;

	size_t space_total, space_end;
	size_t size_chunk_end, size_chunk_start;
	size_t written = 0;

	write_lock_irqsave(&(ringbuf->lock), flags);

	space_total = ringbuf->size - __ringbuf_data_size(ringbuf, 0);
	space_end = ringbuf->size - ringbuf->end;
	
	if (count > ringbuf->size)
		count = ringbuf->size;

	size_chunk_end = MIN(count, space_end);
	size_chunk_start = count - size_chunk_end;
	
	if (memcpy(ringbuf->buffer + ringbuf->end, data, size_chunk_end))
	{
		written += size_chunk_end;

		if (memcpy(ringbuf->buffer, data + size_chunk_end, size_chunk_start))
		{
			written += size_chunk_start;
		}
	}

	ringbuf->end = (ringbuf->end + written) % ringbuf->size;

	if (count > space_total)
		ringbuf->start = (ringbuf->end + 1) % ringbuf->size;

	write_unlock_irqrestore(&(ringbuf->lock), flags);
	
	return written;
}


static size_t __ringbuf_data_read(struct usb_pac7311_ringbuf *ringbuf, unsigned char *data, size_t count, int lock)
{
	unsigned long flags = 0;

	size_t data_total, data_start;
	size_t size_chunk_start, size_chunk_end;
	size_t read = 0;

	if (lock)
	{
		write_lock_irqsave(&(ringbuf->lock), flags);
	}

	data_total = __ringbuf_data_size(ringbuf, 0);
	data_start = ringbuf->size - ringbuf->start;
	
	if (count > data_total)
		count = data_total;
	
	size_chunk_start = MIN(count, data_start);
	size_chunk_end = count - size_chunk_start;
	
	if (memcpy(data, ringbuf->buffer + ringbuf->start, size_chunk_start))
	{
		read += size_chunk_start;

		if (memcpy(data + size_chunk_start, ringbuf->buffer, size_chunk_end))
		{
			read += size_chunk_end;
		}
	}

	ringbuf->start = (ringbuf->start + count) % ringbuf->size;

	if (lock)
	{
		write_unlock_irqrestore(&(ringbuf->lock), flags);
	}
	
	return read;
}


static inline size_t ringbuf_data_read(struct usb_pac7311_ringbuf *ringbuf, unsigned char *data, size_t count)
{
	return __ringbuf_data_read(ringbuf, data, count, 1);
}


static int __ringbuf_data_search(struct usb_pac7311_ringbuf *ringbuf, unsigned char *data, size_t count, size_t offset, int lock)
{
	unsigned long flags = 0;
	
	int retval = -1;
	int data_total, data_start, size_chunk_end, size_chunk_start,
	    abspos, relpos, data_spos, ringbuf_spos;

	if (count <= 0)
	{
		retval = ringbuf->start;
		goto exit0;
	}

	if (lock)
	{	
		read_lock_irqsave(&(ringbuf->lock), flags);
	}
	
	data_total = ((ringbuf->end >= offset)?
	              (ringbuf->end - offset) :
	              (ringbuf->size + ringbuf->end - offset));

	data_total -= (count - 1);
	data_start = ringbuf->size - ringbuf->start;
	
	if (data_total < 0)
	{
		retval = -1;
		goto exit0;
	}
	
	size_chunk_start = MIN(data_total, data_start);
	size_chunk_end = data_total - size_chunk_start;
	
	if (size_chunk_start == 0)
	{
		retval = -1;
		goto exit0;
	}
	
	for (relpos = 0; relpos < size_chunk_start; ++relpos)
	{
		abspos = offset + relpos;
		
		for (data_spos = 0; data_spos < count; ++data_spos)
		{
			ringbuf_spos = (abspos + data_spos) % ringbuf->size;
			
			if (data[data_spos] != ringbuf->buffer[ringbuf_spos])
			{
				break;
			}
		}
		
		if (data_spos >= count)  /* we found a match */
		{
			retval = abspos;
			break;
		}
	}
	
	if (retval >= 0)
	{
		goto exit0;
	}

	if (size_chunk_end == 0)
	{
		retval = -1;
		goto exit0;
	}

	for (relpos = 0; relpos < size_chunk_end; ++relpos)
	{
		abspos = relpos;

		for (data_spos = 0; data_spos < count; ++data_spos)
		{
			ringbuf_spos = abspos + data_spos;
			
			if (data[data_spos] != ringbuf->buffer[ringbuf_spos])
			{
				break;
			}
		}
		
		if (data_spos >= count)  /* we found a match */
		{
			retval = abspos;
			break;
		}
	}


exit0:
	if (lock)
	{
		read_unlock_irqrestore(&(ringbuf->lock), flags);
	}
	
	return retval;
}


static int ringbuf_data_search(struct usb_pac7311_ringbuf *ringbuf, unsigned char *data, size_t count, size_t offset)
{
	return __ringbuf_data_search(ringbuf, data, count, offset, 1);
}


static int ringbuf_data_frame_read(struct usb_pac7311_ringbuf *ringbuf, unsigned char *data, size_t count)
{
	unsigned long flags;
	int frame_start, frame_end, frame_size;

	int retval = 0;

	write_lock_irqsave(&(ringbuf->lock), flags);
	
	frame_start = __ringbuf_data_search(ringbuf, "\xff\xd9", 2, ringbuf->start, 0);
	
	if (frame_start < 0)  /* test if a frame start was found at all */
	{
		retval = -1;
		goto exit0;
	}
	
	frame_start = (frame_start + 2) % ringbuf->size;
	
	frame_end = __ringbuf_data_search(ringbuf, "\xff\xd9", 2, frame_start, 0);
	
	if (frame_end < 0)  /* test if a frame end was found at all */
	{
		retval = -1;
		goto exit0;
	}
	
	frame_end -= 2;

	if (frame_end < 0)
	{
		frame_end += ringbuf->size;
	}
	
	frame_size = ((frame_end >= frame_start)?
	              (frame_end - frame_start) :
	              (ringbuf->size + frame_end - frame_start));

	if (frame_size > count)
	{
		retval = -2;
		goto exit0;
	}
	
	ringbuf->start = frame_start;
	
	retval = __ringbuf_data_read(ringbuf, data, frame_size, 0);
	
exit0:
	write_unlock_irqrestore(&(ringbuf->lock), flags);
	
	return retval;
}

