/*
 * USB PAC7311 Stream Driver - 0.1a
 *
 * Copyright (C) 2006 Jörg Schummer (joerg.schummer@web.de)
 *
 *   This driver allows to receive the data stream from a USB web cam
 *   based on the PixArt VGA Single Chip (PAC7311) via a character
 *   device, namely pac7311_streamX .
 *
 *   Note that the data is not processed in any way, it is only delivered
 *   to user space. This driver has been written only for the purpose of
 *   developing a 'proper' frame decoding driver.
 *
 *   This driver is based on the USB skeleton driver:
 *
 * Copyright (C) 2001-2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/spinlock.h>

#include "usb-pac7311-stream.h"
#include "usb-pac7311-ringbuf.c"


/* Define these values to match your devices */
#define USB_PAC7311_STREAM_VENDOR_ID	0x093a
#define USB_PAC7311_STREAM_PRODUCT_ID	0x2600

/* table of devices that work with this driver */
static struct usb_device_id pac7311_stream_table [] = {
	{ USB_DEVICE(USB_PAC7311_STREAM_VENDOR_ID, USB_PAC7311_STREAM_PRODUCT_ID) },
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, pac7311_stream_table);


/* Get a minor range for your devices from the usb maintainer */
#define USB_PAC7311_STREAM_MINOR_BASE	255


#define PAC7311_ISOC_NOOF_URBS	2	/* no of urbs scheduled at the same time */
#define PAC7311_ISOC_NOOF_PKTS	32	/* max no of packets contained in one urb */

#define PAC7311_READ_BUF_SIZE  65536	/* max bytes of data which can be read from 
					   the pac7311_streamX device at once */
#define PAC7311_READ_EMPTY_TRIES 10	/* number of times empty urbs can return
					   before isoc transfer is aborted */

#define PAC7311_CTRLMSG_TIMEOUT 100	/* the timeout for delivering control messages
					   between the device and this driver, unit: ms */


#define PAC7311_STREAM_DEBUG 1		/* whether to output (a lot of) debug messages */


#if (PAC7311_STREAM_DEBUG == 1)

#define DO_DEBUG(CMDS) \
do { \
	CMDS; \
} while (0);

#else

#define DO_DEBUG(CMDS) do { } while (0);

#endif


/* the PDEBUG macro works as PDEBUG(info, ...) or PDEBUG(err, ...) */
#define PDEBUG(TYPE, MSG, ARGS...) DO_DEBUG(TYPE(MSG, ##ARGS));


/* forward declarations */
static int pac7311_stream_open(struct inode *inode, struct file *file);
static int pac7311_stream_release(struct inode *inode, struct file *file);
static ssize_t pac7311_stream_read(struct file *file, char *buffer, size_t count, loff_t *ppos);
static int pac7311_stream_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int pac7311_stream_probe(struct usb_interface *interface, const struct usb_device_id *id);
static void pac7311_stream_disconnect(struct usb_interface *interface);
static void pac7311_stream_delete(struct kref *kref);


/* usb device operations */
static struct usb_driver pac7311_stream_driver = {
	.owner =	THIS_MODULE,
	.name =	"pac7311_stream",
	.probe =	pac7311_stream_probe,
	.disconnect =	pac7311_stream_disconnect,
	.id_table =	pac7311_stream_table,
};

/* file operations */
static struct file_operations pac7311_stream_fops = {
	.owner =	THIS_MODULE,
	.read =		pac7311_stream_read,
	.ioctl =	pac7311_stream_ioctl,
	.open =		pac7311_stream_open,
	.release =	pac7311_stream_release,
};

/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with devfs and the driver core
 */
static struct usb_class_driver pac7311_stream_class = {
	.name =	"pac7311_stream%d",
	.fops =		&pac7311_stream_fops,
	.minor_base =	USB_PAC7311_STREAM_MINOR_BASE,
};


/* Structure to hold all of our device specific stuff */
struct usb_pac7311_stream {
	struct semaphore		lock;			/* lock for this device. only allow exclusive
								   access at this stage */

	struct usb_device		*udev;			/* the usb device for this device */
	struct usb_interface		*interface;		/* the interface for this device */

	struct urb			*urbs[PAC7311_ISOC_NOOF_URBS];	/* the urbs used for the isoc transfer */
	unsigned char 			*isoc_in_buffer;	/* the buffer used by the urbs */
	size_t				isoc_in_bufsize;	/* the size of the above buffer */
	size_t				isoc_in_pkt_size;	/* the maximum size of a packet */

	__u8				isoc_in_endpointAddr;	/* the address of the isoc in endpoint */
	struct kref			kref;			/* the number of references on this module */

	struct usb_pac7311_ringbuf	*ringbuf;		/* the ringbuffer via which the isoc
								   transfer interrupt and our 'read'
								   device function exchange data */

	int				is_connected;		/* whether the device is connected */
	int				is_streaming;		/* whether the device is currently
								   delivering a data stream */
	wait_queue_head_t		isoc_in_wait_queue;	/* the queue for communication between
								   various events (such as the isoc
								   transfer interrupt) and the 'read'
								   device function */
	spinlock_t			empty_read_lock;	/* lock to protect the empty_read
								   variable */
	int				empty_read;		/* number of empty urbs that have
								   returned since last reset */
	
	unsigned char			*readbuf;		/* the buffer used by 'read' before
								   the data is copied to user-space */
};

#define to_pac7311_stream_dev(d) container_of(d, struct usb_pac7311_stream, kref)

static struct usb_driver pac7311_stream_driver;


/* internal forward declarations */

static int __pac7311_stream_stream_start(struct usb_pac7311_stream *dev);
static int __pac7311_stream_stream_stop(struct usb_pac7311_stream *dev);
static int __pac7311_stream_reg_set(struct usb_pac7311_stream *dev, struct pac7311_stream_reg *reg);
static int __pac7311_stream_reg_get(struct usb_pac7311_stream *dev, struct pac7311_stream_reg *reg);
static int __pac7311_stream_usb_control_msg_send(struct usb_pac7311_stream *dev, __u8 request,
						 __u16 value, __u16 index, char *data, size_t datasize);
static int __pac7311_stream_usb_control_msg_recv(struct usb_pac7311_stream *dev, __u8 request,
						 __u16 value, __u16 index, char *data, size_t datasize);



/******************************************************************************
 *
 * Helper functions
 *
 ******************************************************************************/


static inline void empty_read_inc(struct usb_pac7311_stream *dev)
{
	long flags;

	spin_lock_irqsave(&(dev->empty_read_lock), flags);
	
	++(dev->empty_read);

	spin_unlock_irqrestore(&(dev->empty_read_lock), flags);
}


static inline void empty_read_reset(struct usb_pac7311_stream *dev)
{
	long flags;

	spin_lock_irqsave(&(dev->empty_read_lock), flags);
	
	dev->empty_read = 0;

	spin_unlock_irqrestore(&(dev->empty_read_lock), flags);
}


/******************************************************************************
 *
 * File operation functions
 *
 ******************************************************************************/


/**
 * Isochronous transfer interrupt handler.
 *
 * The data returned by isochronous transfers first arrives here and is then
 * pur into the ringbuffer.
 *
 **/
 
static void pac7311_stream_isoc_handler(struct urb *urb, struct pt_regs *ptregs)
{
	struct usb_pac7311_stream *dev;
	int i;
	
	dev = (struct usb_pac7311_stream *) urb->context;

	
	PDEBUG(info, "urb came back (status %d), %d errors, length %d",
	             urb->status, urb->error_count, urb->actual_length);


	if (urb->actual_length == 0)
	{
		empty_read_inc(dev);
	}


	/* write the data of each packet contained in the urb into the ringbuffer */

	for (i = 0; i < urb->number_of_packets; ++i)
	{
		int len = urb->iso_frame_desc[i].actual_length;  /* the length of the packet */

		if (ringbuf_data_write(dev->ringbuf, urb->transfer_buffer + urb->iso_frame_desc[i].offset,
		                       len) != len)
		{
			err("%s - ringbuffer error", __FUNCTION__);
		}
	}
	

	if (dev->is_connected && dev->is_streaming)  /* only if the device is connected and streaming! */
	{
		if (usb_submit_urb(urb, GFP_ATOMIC))
		{
			err("%s - error submitting urb", __FUNCTION__);

			dev->is_streaming = 0;
				/* don't schedule any more urbs, this way we
				   don't end up having only one URB around, when
				   more should be active */
		}
	}

	wake_up(&(dev->isoc_in_wait_queue));  /* wake up a possible waiting 'read' */
}


/**
 * The device's open function.
 *
 * This function starts streaming the data via isochronous transfers. After a successful
 * call to 'open', the data will be continuously written into the ringbuffer (by the
 * interrupt handler), regardless of whether a 'read' is actually delivering it. This
 * is to maintain a contonuous data flow, even if 'read' is only called once in a while.
 *
 **/

static int pac7311_stream_open(struct inode *inode, struct file *file)
{
	struct usb_pac7311_stream *dev;
	struct usb_interface *interface;
	int subminor;
	int retval;

	retval = 0;

	PDEBUG(info, "pac7311_stream_open called");

	subminor = iminor(inode);

	interface = usb_find_interface(&pac7311_stream_driver, subminor);
	
	if (!interface) {
		err ("%s - error, can't find device for minor %d",
		     __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit0;
	}

	dev = usb_get_intfdata(interface);
	if (!dev) {
		err ("%s - error, can't get interface data for device minor %d",
		     __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit0;
	}

	if (!dev->is_connected) {
		err("%s - error, camera not connected", __FUNCTION__);
		retval = -ENODEV;
		goto exit0;
	}

	if (down_trylock(&(dev->lock))) {  /* at this point, only one user space process can
					      access the stream at the same time */
		err ("%s - error, can't lock device for minor %d",
		     __FUNCTION__, subminor);
		retval = -ENOLCK;
		goto exit0;
	}

	/* increment our usage count for the device */
	kref_get(&dev->kref);

	/* save our object in the file's private structure */
	file->private_data = dev;

exit0:
	return retval;
}


/**
 * The device's release function.
 *
 * This function stops the isochronous transfers and removes
 * all URBs which were used for that.
 *
 **/

static int pac7311_stream_release(struct inode *inode, struct file *file)
{
	struct usb_pac7311_stream *dev;
	int retval, ret;

	retval = 0;

	PDEBUG(info, "pac7311_stream_release called");

	dev = (struct usb_pac7311_stream *)file->private_data;
	if (dev == NULL)
	{
		retval = -ENODEV;
		goto exit0;
	}


	PDEBUG(info, "stopping stream...");

	/* tell the camera to stop streaming, no matter if the 'is_streaming' flag
	   is set. This is to avoid having a closed device, but still have the
	   urbs allocated */

	ret = __pac7311_stream_stream_stop(dev);
		
	if (ret) {
	
		PDEBUG(err, "stopping stream failed.");
		retval = ret;
	}
	else
	{
		PDEBUG(info, "stream stopped.");
	}

	ringbuf_reset(dev->ringbuf);  /* clear the ringbuffer */
	
	up(&(dev->lock));  /* unlock the device, for future access */

	/* decrement the count on our device */
	kref_put(&dev->kref, pac7311_stream_delete);

exit0:
	return retval;
}


/**
 * The device's read function.
 *
 * This function delivers a frame from the ringbuffer to the user space.
 * If there no complete frame in the ringbuffer, it waits for the
 * isochronous transfer interrupt handler to fill the ringbuffer. However,
 * if streaming is stopped or the device is disconnected, while 'read' is
 * called, this wait is omitted and it returns 0, i.e. no data.
 *
 **/
 
static ssize_t pac7311_stream_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	struct usb_pac7311_stream *dev;
	int frame_size = -1;

	dev = (struct usb_pac7311_stream *)file->private_data;

	if (!dev->is_connected)
		return -ENODEV;
		
	if (!dev->is_streaming)
		return -EFAULT;

	/* we only allow a maximum of bytes to be delivered at the same time. This is,
	   because we don't want to allocate a 'read buffer' every time 'read' is
	   called. */

	if (count > PAC7311_READ_BUF_SIZE)
		count = PAC7311_READ_BUF_SIZE;
	
	memset(dev->readbuf, 0, count);


	empty_read_reset(dev);


	PDEBUG(info, "before: start: %d, end: %d", dev->ringbuf->start, dev->ringbuf->end);


	/* As long as we are connected and streaming, but there's no frame in
	   the ring buffer... */

	while (dev->is_connected && dev->is_streaming &&
	       ((frame_size =
	           ringbuf_data_frame_read(dev->ringbuf, dev->readbuf, count)) == -1) &&
	       (dev->empty_read < PAC7311_READ_EMPTY_TRIES))
	{
		/* sleep, until (presumably) the isoc interrupt handler wakes us up.
		   Usually that means that it has put some more data in the ring buffer. */
		   
		interruptible_sleep_on(&(dev->isoc_in_wait_queue));
	}

	PDEBUG(info, "after: start: %d, end: %d", dev->ringbuf->start, dev->ringbuf->end);

	/* next, we can copy the data to user space */

	if (frame_size >= 0)
	{
		frame_size -= copy_to_user(buffer, dev->readbuf, frame_size);
	}
	else
	{
		PDEBUG(info, "ERROR: frame was bigger than buffer provided");
	
		frame_size = 0;
	}

	PDEBUG(info, "%d bytes requested, %d bytes delivered", count, frame_size);

	return frame_size;
}


static int __pac7311_stream_usb_control_msg_send(struct usb_pac7311_stream *dev, __u8 request,
						 __u16 value, __u16 index, char *data, size_t datasize)
{
	int retval;
	unsigned int pipe;
	
/*
extern int usb_control_msg(struct usb_device *dev, unsigned int pipe,
	__u8 request, __u8 requesttype, __u16 value, __u16 index,
	void *data, __u16 size, int timeout);
*/

	pipe = usb_sndctrlpipe(dev->udev, 0);  /* pipe for sending a control message */

	retval = usb_control_msg(dev->udev, pipe, request, USB_DIR_OUT + USB_TYPE_VENDOR + USB_RECIP_DEVICE,
				 value, index, data, datasize, PAC7311_CTRLMSG_TIMEOUT);

	PDEBUG(info, "usb_control_msg returned %d", retval);
	
	return retval;
}


static int __pac7311_stream_usb_control_msg_recv(struct usb_pac7311_stream *dev, __u8 request,
						 __u16 value, __u16 index, char *data, size_t datasize)
{
	int retval;
	unsigned int pipe;
	
/*
extern int usb_control_msg(struct usb_device *dev, unsigned int pipe,
	__u8 request, __u8 requesttype, __u16 value, __u16 index,
	void *data, __u16 size, int timeout);
*/
	
	pipe = usb_rcvctrlpipe(dev->udev, 0);  /* pipe for receiving a control message */

	retval = usb_control_msg(dev->udev, pipe, request, USB_DIR_IN + USB_TYPE_VENDOR + USB_RECIP_DEVICE,
				 value, index, data, datasize, PAC7311_CTRLMSG_TIMEOUT);

	PDEBUG(info, "usb_control_msg returned %d", retval);
	
	return retval;
}


static int __pac7311_stream_stream_start(struct usb_pac7311_stream *dev)
{
	struct urb *urb;
	int pkt_cnt, urb_malloc_cnt, urb_submit_cnt;
	int retval, ret;
	pac7311_stream_regval_t regval;
	
	retval = 0;

	PDEBUG(info, "initialising urbs...");

	/* initialise each of the urbs before they will be submitted later */

	for (urb_malloc_cnt = 0; urb_malloc_cnt < PAC7311_ISOC_NOOF_URBS; ++urb_malloc_cnt)
	{
		urb = dev->urbs[urb_malloc_cnt] = usb_alloc_urb(PAC7311_ISOC_NOOF_PKTS, GFP_KERNEL);
			/* Zeroes the allocated data up to iso_frame_desc[], *not* including the last! */
		
		if (!urb)
		{
			err ("%s - error, can't get memory for urbs",
			     __FUNCTION__);
			
			retval = -ENOMEM;
			goto error0;
		}
		
		urb->dev                    = dev->udev;
		urb->context                = dev;
		urb->pipe                   = usb_rcvisocpipe(dev->udev, dev->isoc_in_endpointAddr);
		urb->transfer_flags         = URB_ISO_ASAP;		/* Leave URB_ASYNC_UNLINK clear and never call usb_unlink_urb() - ? */
		urb->complete               = pac7311_stream_isoc_handler;
		urb->number_of_packets      = PAC7311_ISOC_NOOF_PKTS;
		urb->transfer_buffer        = dev->isoc_in_buffer;
		urb->transfer_buffer_length = dev->isoc_in_bufsize;
		urb->interval               = 1;			/* See Table 9-10 of the USB 1.1 specification */
		
		/* set each single packet's buffer offset and maximum length */
		
		for (pkt_cnt = 0; pkt_cnt < PAC7311_ISOC_NOOF_PKTS; ++pkt_cnt) {
			urb->iso_frame_desc[pkt_cnt].offset =
				(urb_malloc_cnt * PAC7311_ISOC_NOOF_PKTS * dev->isoc_in_pkt_size) +
				(pkt_cnt * dev->isoc_in_pkt_size);
			urb->iso_frame_desc[pkt_cnt].length = dev->isoc_in_pkt_size;
		}
	}

	PDEBUG(info, "urbs initialised.");

	PDEBUG(info, "starting stream...");

	ret = __pac7311_stream_usb_control_msg_recv(dev, 0x00, 0x00, 0x78, &regval, 1);
	
	if (ret == 1)
	{
		regval |= 0x01;
	
		ret = __pac7311_stream_usb_control_msg_send(dev, 0x00, regval, 0x78, &regval, 1);
	}

	if (ret != 1) {  /* abort, if starting the stream failed */
		err("%s - starting stream failed", __FUNCTION__);
		
		retval = ret;
		goto error0;
	}
	else
	{
		PDEBUG(info, "stream started.");
	}


	/* prepare for streaming. Now everything has to be brought into a state such
	   that our isochronous transfer handler can do its job */

	ringbuf_reset(dev->ringbuf);

	dev->is_streaming = 1;


	ret = 0;

	PDEBUG(info, "submitting urbs (ret = %d)...", ret);

	/* submit the previously initialised urbs */

	for (urb_submit_cnt = 0; (urb_submit_cnt < PAC7311_ISOC_NOOF_URBS) && (!retval); ++urb_submit_cnt)
	{
		PDEBUG(info, "submitting urb no %d...", urb_submit_cnt);

		ret = usb_submit_urb(dev->urbs[urb_submit_cnt], GFP_KERNEL);

		PDEBUG(info, "urb no %d submitted (ret = %d).", urb_submit_cnt, ret);
	}

	if (ret)  /* abort, if (one of) the urbs could not be submitted */
	{
		err ("%s - error, can't submit urbs",
		     __FUNCTION__);

		retval = -EFAULT;
		goto error1;
	}
	else
	{
		PDEBUG(info, "urbs submitted.");
	}

	goto exit0;

error1:  /* clean up: stop streaming */
	dev->is_streaming = 0;

	while (--urb_submit_cnt > 0)
	{
		usb_kill_urb(dev->urbs[urb_malloc_cnt]);
	}     

error0:  /* clean up: free urbs' memory */
	while (--urb_malloc_cnt > 0)
	{
		usb_free_urb(dev->urbs[urb_malloc_cnt]);
		dev->urbs[urb_malloc_cnt] = NULL;
	}
	
exit0:
	return retval;
}

static int __pac7311_stream_stream_stop(struct usb_pac7311_stream *dev)
{
	int i, ret = 0, retval = 0;
	pac7311_stream_regval_t regval;

	if (!dev->is_connected)
	{
		err("%s - camera is not connected", __FUNCTION__);

		retval = -ENODEV;
		goto exit0;
	}

	/* stop streaming by unsetting the LSB of register 0x78 */

	ret = __pac7311_stream_usb_control_msg_recv(dev, 0x00, 0x00, 0x78, &regval, 1);
	
	if (ret == 1)
	{
		regval &= 0xfe;
	
		ret = __pac7311_stream_usb_control_msg_send(dev, 0x00, regval, 0x78, &regval, 1);
	}


	if (ret != 1)
	{
		retval = -EFAULT;
	}
	
exit0:
	dev->is_streaming = 0;  /* mark that we stopped streaming */

	PDEBUG(info, "deleting urbs...");

	/* delete and free all urbs */
	
	for (i = 0; i < PAC7311_ISOC_NOOF_URBS; ++i)
	{
		usb_kill_urb(dev->urbs[i]);
		usb_free_urb(dev->urbs[i]);
		dev->urbs[i] = NULL;
	}
	
	PDEBUG(info, "urbs deleted.");

	return retval;
}

static inline int __pac7311_stream_reg_set(struct usb_pac7311_stream *dev, struct pac7311_stream_reg *reg_user)
{
	int retval = 0;
	struct pac7311_stream_reg *reg;
	
	if (!(reg = kmalloc(sizeof(struct pac7311_stream_reg), GFP_KERNEL)))
	{
		retval = -ENOMEM;
		goto exit0;
	}
	
	if (copy_from_user(reg, reg_user, sizeof(struct pac7311_stream_reg)))
	{
		retval = -EFAULT;
		goto exit1;
	}
	
	if (__pac7311_stream_usb_control_msg_send(dev, 0x00, reg->val, reg->reg, &(reg->val), 1) != 1)
	{
		retval = -EFAULT;
		goto exit1;
	}

exit1:
	kfree(reg);

exit0:	
	return retval;
}

static inline int __pac7311_stream_reg_get(struct usb_pac7311_stream *dev, struct pac7311_stream_reg *reg_user)
{
	int retval = 0;
	struct pac7311_stream_reg *reg;
	
	if (!(reg = kmalloc(sizeof(struct pac7311_stream_reg), GFP_KERNEL)))
	{
		retval = -ENOMEM;
		goto exit0;
	}
	

	if (copy_from_user(reg, reg_user, sizeof(struct pac7311_stream_reg)))
	{
		retval = -EFAULT;
		goto exit1;
	}
	

	if (__pac7311_stream_usb_control_msg_recv(dev, 0x00, 0x00, reg->reg, &(reg->val), 1) != 1)
	{
		retval = -EFAULT;
		goto exit1;
	}


	if (copy_to_user(reg_user, reg, sizeof(struct pac7311_stream_reg)))
	{
		retval = -EFAULT;
		goto exit1;
	}

exit1:
	kfree(reg);
	
exit0:
	return retval;
}


static inline int __pac7311_stream_control_msg_send(struct usb_pac7311_stream *dev, struct pac7311_stream_control_msg *msg_user)
{
	int ret, retval = 0;
	struct pac7311_stream_control_msg *msg;
		
	if (!(msg = kmalloc(sizeof(struct pac7311_stream_control_msg), GFP_KERNEL)))
	{
		retval = -ENOMEM;
		goto exit0;
	}

	if (copy_from_user(msg, msg_user, sizeof(struct pac7311_stream_control_msg)) != 0)
	{
		retval = -EFAULT;
		goto exit1;
	}

	if (!(msg->data = kmalloc(msg->len * sizeof(char), GFP_KERNEL)))
	{
		retval = -ENOMEM;
		goto exit1;
	}
	
	if (copy_from_user(msg->data, msg_user->data, msg->len * sizeof(char)))
	{
		retval = -EFAULT;
		goto exit2;
	}


	DO_DEBUG(
		char *asc_msg = kmalloc(msg->len * sizeof(char) * 2 + 1, GFP_KERNEL);
		char cur;				
		int i;
		
		if (asc_msg)
		{
			for (i = 0; i < msg->len; ++i)
			{
				cur = ((msg->data[i] >> 4) & 0x0f);
				asc_msg[i * 2 + 0] = (cur < 0xa)? (cur + '0'):(cur - 0xa + 'A');

				cur = (msg->data[i] & 0x0f);
				asc_msg[i * 2 + 1] = (cur < 0xa)? (cur + '0'):(cur - 0xa + 'A');
			}
			
			asc_msg[i * 2] = '\0';

			PDEBUG(info, "calling __pac7311_stream_usb_control_msg_send with index=0x%x, data=%s, len=0x%x", msg->index, asc_msg, msg->len);
			
			kfree(asc_msg);	
		}
	);
	
	ret = __pac7311_stream_usb_control_msg_send(dev, 0x00, 0x00, msg->index, msg->data, msg->len);
	
	PDEBUG(info, "__pac7311_stream_usb_control_msg_send returned %d", ret);
	
	if (ret != msg->len)
	{
		retval = -EFAULT;
		goto exit2;
	}


exit2:
	kfree(msg->data);

exit1:
	kfree(msg);

exit0:
	return retval;
}


static int pac7311_stream_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct usb_pac7311_stream *dev;
	int retval = 0, access_error = 0;


	PDEBUG(info, "%s called", __FUNCTION__);
	

	if ((_IOC_TYPE(cmd) != PAC7311_STREAM_IOCTL_MAGIC) ||
	    (_IOC_NR(cmd) > PAC7311_STREAM_IOCTL_MAX_NR))
	{
		PDEBUG(err, "%s - wrong command", __FUNCTION__);
		return -ENOTTY;
	}


	if ((_IOC_DIR(cmd) & _IOC_READ) &&
	    (!access_ok(VERIFY_WRITE, (void *) arg, _IOC_SIZE(cmd))))
	{
		PDEBUG(err, "%s - could not get write access", __FUNCTION__);
		access_error = 1;
	}
	if ((_IOC_DIR(cmd) & _IOC_WRITE) &&
	    (!access_ok(VERIFY_READ, (void *) arg, _IOC_SIZE(cmd))))
	{
		PDEBUG(err, "%s - could not get read access", __FUNCTION__);
		access_error = 1;
	}
	
	if (access_error)
		return -EFAULT;
	

	dev = (struct usb_pac7311_stream *)file->private_data;
	if ((dev == NULL) || (!dev->is_connected))
	{
		return -ENODEV;
	}


	switch (cmd)
	{
	case PAC7311_STREAM_IOCTL_STREAM_START:
		retval = __pac7311_stream_stream_start(dev);
		break;
		
	case PAC7311_STREAM_IOCTL_STREAM_STOP:
		retval = __pac7311_stream_stream_stop(dev);
		break;
		
	case PAC7311_STREAM_IOCTL_REG_SET:
		retval = __pac7311_stream_reg_set(dev, (struct pac7311_stream_reg *) arg);
		break;
		
	case PAC7311_STREAM_IOCTL_REG_GET:
		retval = __pac7311_stream_reg_get(dev, (struct pac7311_stream_reg *) arg);
		break;
		
	case PAC7311_STREAM_IOCTL_CONTROL_MSG_SEND:
		retval = __pac7311_stream_control_msg_send(dev, (struct pac7311_stream_control_msg *) arg);
		break;

	default:
		return -ENOTTY;
	}
	
	return retval;
}


/******************************************************************************
 *
 * USB device operation functions
 *
 ******************************************************************************/

/**
 * The device's probe function.
 *
 * Prepares everything for streaming, i.e. finds the device's isochrononous endpoint,
 * allocates the necessary memory, and initialises the ring buffer.
 *
 **/

static int pac7311_stream_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_pac7311_stream *dev = NULL;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;  /* the isochronous endpoint to be found */
	size_t buffer_size;
	int i;
	int retval, ret;
	
	retval = 0;

	/* allocate memory for our device state and initialize it */
	dev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		err("Out of memory");
		retval = -ENOMEM;
		goto exit0;
	}

	memset(dev, 0x00, sizeof(*dev));

	/* just for good style - initialise all the urbs to zero. This
	   is done because none-zero urbs are freed when we close the device
	   or we are disconnected */
	   
	for (i = 0; i < PAC7311_ISOC_NOOF_URBS; ++i)
	{
		dev->urbs[i] = NULL;
	}

	/* allocate memory for the buffer to be used by the 'read' function */
	
	if (!(dev->readbuf = kmalloc(PAC7311_READ_BUF_SIZE, GFP_KERNEL)))
	{
		err("Out of memory");
		retval = -ENOMEM;
		goto error0;
	}

	/* let's set the isoc_in_buffer entry to NULL (just for good style), we
	   might need that in pac7311_stream_delete if the above memset were gone. */
	dev->isoc_in_buffer = NULL;

	/* initialise the kernel reference */
	kref_init(&dev->kref);

	/* get our usb device */
	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;

	/* set alternative interface 8, this enables full transfer speed */
	if (usb_set_interface(dev->udev, 0, 8) < 0) {

		err("Could not set usb interface!");
		retval = -ENODEV;
		goto error0;
	}

	iface_desc = interface->cur_altsetting;

	/* set up the endpoint information, find the first isochronous-in endpoint */
	
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
	
		endpoint = &iface_desc->endpoint[i].desc;  /* the current endpoint */

		/* if the current endpoint is an isochronous 'in' endpoint */

		if (((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
					== USB_DIR_IN) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_ISOC)) {

			dev->isoc_in_pkt_size = le16_to_cpu(endpoint->wMaxPacketSize);

			/* we use one buffer for all urbs */
			buffer_size = PAC7311_ISOC_NOOF_PKTS * PAC7311_ISOC_NOOF_URBS * dev->isoc_in_pkt_size;

			/* initialise the buffer where the isoc urbs' data is stored */			
			dev->isoc_in_bufsize = buffer_size;
			dev->isoc_in_buffer = kmalloc(buffer_size, GFP_KERNEL);

			if (!(dev->isoc_in_buffer)) {
				err("Could not allocate isoc_in_buffer");
				retval = -ENOMEM;
				goto error0;
			}
			
			dev->isoc_in_endpointAddr = endpoint->bEndpointAddress;
			
			break;
		}
	}

	/* if we didn't find an isochronous 'in' endpoint */
	if (!(dev->isoc_in_endpointAddr)) {
		err("Could not find isochronous in-endpoint");
		retval = -ENODEV;
		goto error0;
	}

	/* initialise the ring buffer. At this point, it has been decided
	   that twice the urbs' buffer size provides enough space */
	if (!ringbuf_init(&(dev->ringbuf), dev->isoc_in_bufsize * 2))
	{
		err("Could not allocate ringbuffer");
		retval = -ENOMEM;
		goto error0;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* we can register the device now, as it is ready */
	ret = usb_register_dev(interface, &pac7311_stream_class);
	if (ret) {
		/* something prevented us from registering this driver */
		err("Not able to get a minor for this device.");
		retval = -ENODEV;
		goto error0;
	}

	init_MUTEX(&(dev->lock));  /* initialise the mutex for accessing the device */

	/* init the waitqueue for notification communication between
	   the interrupt and the 'read' function */
	init_waitqueue_head(&(dev->isoc_in_wait_queue));

	spin_lock_init(&(dev->empty_read_lock));

	dev->is_connected = 1;  /* now we dare say the device is conntected */

	/* let the user know what node this device is now attached to */
	info("USB PAC7311 now attached to device minor %d", interface->minor);

	goto exit0;

error0:
	if (dev)
		kref_put(&dev->kref, pac7311_stream_delete);

exit0:
	return retval;
}


/**
 * The device's disconnect function.
 *
 * Tells all other components that we are disconnected so
 * they can stop streaming and delivering data.
 *
 **/

static void pac7311_stream_disconnect(struct usb_interface *interface)
{
	struct usb_pac7311_stream *dev;
	int minor = interface->minor;

	/* prevent pac7311_stream_open() from racing pac7311_stream_disconnect() */
	lock_kernel();

	PDEBUG(info, "disconnecting...");

	dev = usb_get_intfdata(interface);
	
	if (!dev->is_connected)
	{
		PDEBUG(err, "trying to disconnect a disconnected device!");
	}

	PDEBUG(info, "%s: telling read to stop...", __FUNCTION__);

	/* give pac7311_stream_read a chance to get out of this... */
	dev->is_connected = 0;
	wake_up(&(dev->isoc_in_wait_queue));
	
	PDEBUG(info, "%s: stopped read.", __FUNCTION__);

	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &pac7311_stream_class);

	unlock_kernel();

	/* decrement our usage count */
	kref_put(&dev->kref, pac7311_stream_delete);

	info("USB PAC7311 #%d now disconnected", minor);
}


/**
 * The device's delete function.
 *
 * Cleans up all memory used by the driver for a specific device.
 *
 **/
 
static void pac7311_stream_delete(struct kref *kref)
{	
	struct usb_pac7311_stream *dev = to_pac7311_stream_dev(kref);

	PDEBUG(info, "%s - deleting...", __FUNCTION__);

	ringbuf_destroy(dev->ringbuf);

	/* if space is allocated to the urb buffer */
	if (dev->isoc_in_buffer)
	{
		/* free it */
		kfree(dev->isoc_in_buffer);
		dev->isoc_in_buffer = NULL;
	}

	/* if space is allocated to the read buffer */
	if (dev->readbuf)
	{
		/* free it */
		kfree(dev->readbuf);
		dev->readbuf = NULL;
	}

	usb_put_dev(dev->udev);

	/* free our device storage space */
	kfree(dev);

	PDEBUG(info, "%s - deleted.", __FUNCTION__);
}


/******************************************************************************
 *
 * Kernel driver functions
 *
 ******************************************************************************/

static int __init usb_pac7311_stream_init(void)
{
	int result;

	/* register this driver with the USB subsystem */
	result = usb_register(&pac7311_stream_driver);
	if (result)
		err("usb_register failed. Error number %d", result);

	return result;
}

static void __exit usb_pac7311_stream_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&pac7311_stream_driver);
}


module_init (usb_pac7311_stream_init);
module_exit (usb_pac7311_stream_exit);


MODULE_DESCRIPTION("Stream Driver for PixArt VGA Single Chip (PAC7311) based USB webcams");
MODULE_VERSION("0.1a");
MODULE_AUTHOR("Jörg Schummer");
MODULE_LICENSE("GPL");
